#include <rxduino.h>
#include <Ethernet.h>
#include "HTTPClient.h"

// #define DEBUG
#ifdef DEBUG
#include <stdio.h>
#endif

static EthernetClient client;

static bool parseURL(const char *url, char *domain, int *port, char *path ) ;
static bool getIP(const char *domain, byte **ipaddr) ;
static bool sockConn(byte *ipaddr, int port) ;
static  HTTPResult httpReq(HTTP_METH meth, const char *url, const char *dataOut, char *dataIn, int len, int timeout) ;
static bool recieveProcess(char *buff, int len) ; 
static HTTPResult getRetCode(char *dataIn) ;
static bool waitDisconn() ;

static void postByHttp(char *host, char *path, const char *data, int len) ;
static void getByHttp (char *host, char *path) ;

HTTPClient::HTTPClient() { }

HTTPResult HTTPClient::get(const char* url, char* result, int maxResultLen, int timeout) { //Blocking
    return(
        httpReq(HTTP_GET, url,"", result, maxResultLen, timeout) 
        ) ;

}

HTTPResult HTTPClient::post(const char* url, const char *dataOut, char *dataIn, int timeout) { //Blocking
    return(
#    define MAXRETURN 200
        httpReq(HTTP_POST, url, dataOut, dataIn, MAXRETURN, timeout) 
        ) ;
}

static  HTTPResult httpReq(HTTP_METH meth, const char* url, const char *dataOut, char *dataIn, int maxlen, int timeout) {
#define DOMAINSIZE 100
char domain[DOMAINSIZE] ;
int port ;
#define PATHSIZE 200
char path[PATHSIZE] ;
#define IPv4 4
byte **ipaddr ;

    if(!parseURL(url, domain, &port, path))
        return(HTTP_PARSE) ;

#   define RETRY_CONN 5
    for(int i=0; i<RETRY_CONN; i++) {
        if(!getIP(domain, ipaddr))
            return(HTTP_DNS) ;
        if(sockConn(*ipaddr, port))
            goto HTTPreq ;
        delay(100) ;
    }
    return(HTTP_CONN) ; // connection retry out

HTTPreq:
    switch(meth) {
    case HTTP_POST:
        postByHttp(domain, path, dataOut, strlen(dataOut)) ;
        break ;
    case HTTP_GET:
        getByHttp(domain, path) ;
        break ;
    default : return(HTTP_ERROR) ;
    }

    while(!recieveProcess(dataIn, maxlen)) 
#       ifdef DEBUG
        printf("Retry Response\n") 
#       endif
        ; 
#   ifdef DEBUG
    printf("Return====\n%s\n===\n", dataIn) ;    
#   endif
    HTTPResult ret = getRetCode(dataIn) ;
    
    while(!waitDisconn()) 
        printf("Retry disconnectr\n")
        ; 
    delay(100) ;
    return(ret) ;
}

static bool parseURL(const char *url, char *domain, int *port, char *path) {
#define DEFAULTPORT 80
#define MAXURL 400
bool ret ;
int h = 0 ;

    if(strlen(url) >= MAXURL)return(false) ;
    
    *port = DEFAULTPORT ;
    ret = true ;
    path[0] = '/' ; path[1] = '\0' ; // Assume no path
    
    for(int i = 0; ; i++) {
        domain[i] = url[i] ;
        if(url[i] == '\0') {
            break ;
        } 
        if(url[i] == ':') {
            domain[i] = '\0' ;
            i ++ ;
            *port = 0 ;
#           define MAXPORTDIGIT 5
            for(int j=0; j<MAXPORTDIGIT; i++, j++) { // get port
                if((url[i] < '0') || (url[i] > '9'))
                    break ;
                *port *= 10 ;  
                *port += (url[i] - '0') ;       
            }
#           ifdef DEBUG
            printf("Port=%d\n", *port) ;
#           endif
            if(url[i] == '\0')
                break ;
        }
        if(url[i] == '/') {
            domain[i] = '\0' ;
            int j ;
            for(j=0; j<PATHSIZE; i++, j++) { // get /path
                path[j] = url[i] ;
                if(url[i] == '\0') break ;
            }
#           ifdef DEBUG
            printf("Path=%s\n", path) ;
#           endif
            if(j==PATHSIZE)ret = false ;
            break ;           
        } 
    }

#   ifdef DEBUG
    printf("Input:%s\n", url) ;
    printf("Result:%d, Domain:%s, Port:%d, Path:%s\n", ret, domain, *port, path) ;
#   endif
    return ret ;
}

#ifdef DEBUG
static void test_get() {
// #define BUFFSIZE 100
char domain[BUFFSIZE] ;
int port ;
char path[BUFFSIZE] ;

    parseURL("abc.bcd.com", domain, &port, path) ;
    parseURL("abc.bcd.com:123", domain, &port, path) ;    
    parseURL("abc.bcd.com:1234", domain, &port, path) ;
    parseURL("abc.bcd.com:12345", domain, &port, path) ;    
    parseURL("abc.bcd.com:123456", domain, &port, path) ; 
    parseURL("abc.bcd.com:12/abc", domain, &port, path) ;
    parseURL("abc.bcd.com:123/abc/bcd", domain, &port, path) ;
    parseURL("abc.bcd.com:1234/abc/bcd", domain, &port, path) ;
    parseURL("abc.bcd.com:123456/abc/bcd", domain, &port, path) ;
    parseURL("abc.bcd.com:123/", domain, &port, path) ;
    parseURL("abc.bcd.com/abc", domain, &port, path) ;
    parseURL("abc.bcd.com/", domain, &port, path) ;
}
#endif

static bool getIP(const char *domain, byte **ipaddr) {

    *ipaddr = Ethernet.gethostbyname(domain) ;
    if(((*ipaddr[0])+(*ipaddr[1])+(*ipaddr)[2]+(*ipaddr)[3]) != 0) {
#       ifdef DEBUG
        printf("IP=%d.%d.%d.%d\n", (*ipaddr)[0], (*ipaddr)[1], (*ipaddr)[2], (*ipaddr)[3] ) ;
#       endif
        return(true) ;
    } else {
#       ifdef DEBUG
        printf("ERROR: get IP address (Domain = %s)\n", domain) ;
#       endif

        return(false) ;
    }
}

static bool sockConn(byte *ipaddr, int port) {
char itoa_buff[10] ;

#       ifdef DEBUG
        printf("client.connect(%d.%d.%d.%d, %d)\n", ipaddr[0],ipaddr[1],ipaddr[2],ipaddr[3], port) ;
#       endif
        if (client.connect(ipaddr, port)) {
#            ifdef DEBUG
             Serial.println("connecting...");
#            endif
             return(true) ;
         } else {
             // if you didn't get a connection to the server:
             // Serial.println("connection failed");
#            ifdef DEBUG
             printf("Retry connect\n") ;
#            endif
             client.stop() ;
             // delay(300) ;
         }
 
#       ifdef DEBUG
        printf("Connection Failed\n") ;        
#       endif
        return(false) ;
}

static void postByHttp(char *host, char *path, const char *data, int len) {
#define HTTP_BUFFSIZE (MAXURL+100)
static char httpbuff[HTTP_BUFFSIZE] ;
static char itoa_buff[10] ;

    httpbuff[0] = '\0' ;
    strcat(httpbuff, "POST ") ;
    strcat(httpbuff, path) ;
    strcat(httpbuff, " HTTP/1.1\n") ;
    strcat(httpbuff, "Host: ") ;
    strcat(httpbuff, host) ;
    strcat(httpbuff, "\n") ;
    strcat(httpbuff, "User-Agent: arduino-ethernet\n");
    strcat(httpbuff, "Connection: Close\n");
    strcat(httpbuff, "Content-Length: ");
    sprintf(itoa_buff,"%d", len) ;
    strcat(httpbuff, itoa_buff) ;
    strcat(httpbuff, "\n\n") ;
#   ifdef DEBUG
    printf("Req=%s\n",httpbuff) ;
#   endif
    client.print(httpbuff);
    client.print(data);
#   ifdef DEBUG
    printf("POST requested\n") ;
#   endif
    return ;
}

static void getByHttp (char *host, char *path) {
#define HTTP_BUFFSIZE (MAXURL+100)
static char httpbuff[HTTP_BUFFSIZE] ;
static char itoa_buff[10] ;

    httpbuff[0] = '\0' ;
    strcat(httpbuff, "GET ") ;
    strcat(httpbuff, path) ;
    strcat(httpbuff, " HTTP/1.1\n") ;
    strcat(httpbuff, "Host: ") ;
    strcat(httpbuff, host) ;
    strcat(httpbuff, "\n") ;
    strcat(httpbuff, "User-Agent: arduino-ethernet\n");
    strcat(httpbuff, "Connection: Close\n");
#   ifdef DEBUG
    printf("HTTP Req===>\n%s\n",httpbuff) ;
#   endif
    client.print(httpbuff);
    client.println() ;
    
#   ifdef DEBUG
    printf("POST requested\n") ;
#   endif
    return ;
}

static bool recieveProcess(char *dataIn, int len){
char *p ;

    p = dataIn ;
    do {
        Ethernet.processPackets();
        while (client.available()) {
            *p = client.read();
            if(p >= dataIn+len) {
                // printf("Buffer Overflow %d\n", len) ;
                break ; // just wait for disconnection
            } else p ++ ;
        }
        *p = '\0' ;     
      } while(client.connected()) ;
      return(true) ;
}

static HTTPResult getRetCode(char *dataIn) {
#define ATOI(c) (c - '0')
int i ;
    for(i=0 ; dataIn[i] != ' '; i++) ; // skip to 1st blank
    i++ ;
    // return next 3 digits as int code
    switch(ATOI(dataIn[i])*100 + ATOI(dataIn[i+1])*10+ATOI(dataIn[i+2])) {
    case 404: return(HTTP_NOTFOUND) ; ///<HTTP 404 Error
    case 405: return(HTTP_REFUSED) ; ///<HTTP 403 Error
    case 200: return(HTTP_OK) ;
    default: return(HTTP_ERROR) ; //<HTTP xxx error
    }
}

static bool waitDisconn() {
  // if the server's disconnected, stop the client:
  if (!client.connected()) {
#       ifdef DEBUG
        Serial.println();
        Serial.println("disconnecting.");
#       endif
        client.stop();
        return(true) ;
    } 
    return(false) ;
}

